import pyark.cva_client as cva_client
import pandas as pd


class EntitiesClient(cva_client.CvaClient):

    def __init__(self, url_base, token):
        cva_client.CvaClient.__init__(self, url_base, token=token)

    def get_panels_summary(self, as_data_frame=False, **params):
        """
        :param as_data_frame: return results in a flattened Pandas data frame or in a list of dictionaries
        :type as_data_frame: bool
        :return: returns all observed panels and the number of cases on which they were applied.
        :rtype: list or pd.DataFrame
        """
        results, _ = self._get("panels", **params)
        # some additional flattening
        return self._render(results, as_data_frame=as_data_frame)

    def get_panels_by_regex(self, regex, as_data_frame=False, **params):
        """
        :param regex: the regex query to perform a search
        :type regex: str
        :param as_data_frame: return results in a flattened Pandas data frame or in a list of dictionaries
        :type as_data_frame: bool
        :return: returns observed panels matching the regex.
        :rtype: list or pd.DataFrame
        """
        params['regex'] = regex
        results, _ = self._get("panels/search", **params)
        return self._render(results, as_data_frame=as_data_frame)

    def get_all_panels(self, **params):
        """
        :return: return a list of observed panel names
        :rtype: pd.Series
        """
        results = self.get_panels_summary(consider_versions=False, **params)
        all_panels = [x['panel']['panelName'] for x in results]
        return pd.Series(all_panels, index=all_panels)

    def get_disorders_summary(self, as_data_frame=False, **params):
        """
        :param as_data_frame: return results in a flattened Pandas data frame or in a list of dictionaries
        :type as_data_frame: bool
        :return:
        :rtype: list or pd.DataFrame
        """
        results, _ = self._get("disorders", **params)
        return self._render(results, as_data_frame=as_data_frame)

    def get_disorders_by_regex(self, as_data_frame=False, **params):
        """
        :param as_data_frame: return results in a flattened Pandas data frame or in a list of dictionaries
        :type as_data_frame: bool
        :return: returns observed disorders matching the regex.
        :rtype: list or pd.DataFrame
        """
        results, _ = self._get("disorders/search", **params)
        return self._render(results, as_data_frame=as_data_frame)

    def get_all_specific_diseases(self, **params):
        """
        :return: return a list of observed specific diseases
        :rtype: pd.Series
        """
        results = self.get_disorders_summary(**params)
        all_diseases = list(set([x['disorder']['specificDisease'] for x in results]))
        return pd.Series(all_diseases, index=all_diseases)

    def get_all_disease_groups(self, **params):
        """
        :return: return a list of observed disease groups
        :rtype: pd.Series
        """
        results = self.get_disorders_summary(**params)
        all_disease_groups = list(set([x['disorder']['diseaseGroup'] for x in results]))
        return pd.Series(all_disease_groups, index=all_disease_groups)

    def get_all_disease_subgroups(self, **params):
        """
        :return: return a list of observed disease subgroups
        :rtype: pd.Series
        """
        results = self.get_disorders_summary(**params)
        all_disease_subgroups = list(set([x['disorder']['diseaseSubGroup'] for x in results]))
        return pd.Series(all_disease_subgroups, index=all_disease_subgroups)

    def get_genes_summary(self, as_data_frame=False, **params):
        """
        :param as_data_frame: return results in a flattened Pandas data frame or in a list of dictionaries
        :type as_data_frame: bool
        :return:
        :rtype: list or pd.DataFrame
        """
        results, _ = self._get("genes", **params)
        return self._render(results, as_data_frame=as_data_frame)

    def get_genes(self, as_data_frame=False, **params):
        """
        :param as_data_frame: return results in a flattened Pandas data frame or in a list of dictionaries
        :type as_data_frame: bool
        :return:
        """
        results, _ = self._get(endpoint="genes/search", **params)
        return self._render(results, as_data_frame=as_data_frame)

    def get_phenotypes(self, as_data_frame=False, **params):
        """
        :param as_data_frame: return results in a flattened Pandas data frame or in a list of dictionaries
        :type as_data_frame: bool
        :return:
        """
        results, _ = self._get(endpoint="phenotypes", **params)
        return self._render(results, as_data_frame=as_data_frame)

    def get_hpo(self, identifier, as_data_frame=False):
        """
        :param identifier: An HPO identifier as in HP:00012345
        :type identifier: dict
        :param as_data_frame: return results in a flattened Pandas data frame or in a list of dictionaries
        :type as_data_frame: bool
        :return:
        :rtype: list or pd.DataFrame
        """
        results, _ = self._get("hpos/{id}".format(id=identifier))
        return self._render(results, as_data_frame=as_data_frame)

    def get_hpos(self, as_data_frame=False, max_results=None, **params):
        """
        :param as_data_frame: return results in a flattened Pandas data frame or in a list of dictionaries
        :type as_data_frame: bool
        :type max_results: int
        :return:
        """
        return self._paginate(endpoint="hpos/search", as_data_frame=as_data_frame, max_results=max_results, **params)

    def get_organisations(self, as_data_frame=False, max_results=None, **params):
        """
        :param as_data_frame: return results in a flattened Pandas data frame or in a list of dictionaries
        :type as_data_frame: bool
        :type max_results: int
        :return:
        """
        return self._paginate(endpoint="organisations", as_data_frame=as_data_frame, max_results=max_results, **params)



import pyark.cva_client as cva_client
from protocols.protocol_7_2.cva import ReportEventType, Transaction
import logging
import pandas as pd


REPORT_EVENT_TYPES = [ReportEventType.genomics_england_tiering, ReportEventType.candidate, ReportEventType.reported,
                      ReportEventType.questionnaire]


class CasesClient(cva_client.CvaClient):

    _BASE_ENDPOINT = "cases"

    def __init__(self, url_base, token):
        cva_client.CvaClient.__init__(self, url_base, token=token)

    def count(self, **params):
        """
        :type params: dict
        :rtype: int
        """
        params['count'] = True
        return self.get_cases(**params)

    def get_cases_ids(self, as_data_frame=False, max_results=None, **params):
        """
        :type as_data_frame: bool
        :type max_results: int
        :type params: dict
        :rtype: generator
        """
        params['include'] = ["identifier", "version"]
        return self._paginate(
            endpoint=self._BASE_ENDPOINT, as_data_frame=as_data_frame, max_results=max_results,
            transformer=lambda x: "{}-{}".format(x["identifier"], x["version"]), **params)

    def get_cases(self, as_data_frame=False, max_results=None, include_all=True, **params):
        """
        :type as_data_frame: bool
        :type max_results: int
        :param include_all: use False for the default minimal representation of case, it will be faster
        :type include_all: bool
        :type params: dict
        :rtype: generator
        """
        if params.get('count', False):
            results, next_page_params = self._get(self._BASE_ENDPOINT, **params)
            return results[0]
        else:
            if include_all:
                params['include'] = [self._INCLUDE_ALL]
            return self._paginate(
                endpoint=self._BASE_ENDPOINT, as_data_frame=as_data_frame, max_results=max_results, **params)

    def get_summary(self, as_data_frame=False, params_list=[], **params):
        """
        :type as_data_frame: bool
        :type params_list: list
        :rtype: dict | pd.DataFrame
        """
        if params_list:
            self._params_sanity_checks(params_list)
            for p in params_list:
                p.update(params)
            results_list = [self.get_summary(as_data_frame=as_data_frame, **p) for p in params_list]
            return self._render_multiple_results(results_list, as_data_frame=as_data_frame)
        else:
            results, _ = self._get("{endpoint}/summary".format(endpoint=self._BASE_ENDPOINT), **params)
            if not results:
                logging.warning("No summary found")
                return None
            assert len(results) == 1, "Unexpected number of summaries"
            return self._render_single_result(results, as_data_frame=as_data_frame, indexes=params)

    def delete(self, case_id, case_version):
        path = "{endpoint}/{case_id}/{case_version}".format(
            endpoint=self._BASE_ENDPOINT, case_id=case_id, case_version=case_version
        )
        results, _ = self._delete(path)
        result = self._render_single_result(results)
        return Transaction.fromJsonDict(result) if result else None

    @staticmethod
    def _params_sanity_checks(params_list):
        if not all(isinstance(p, dict) for p in params_list):
            raise ValueError("Cannot accept a list of 'params' combined with other parameters. " +
                             "Include all parameters in the list")
        keys = None
        for p in params_list:
            if keys is None:
                keys = set(p.keys())
            else:
                if len(set(p.keys()).difference(keys)) > 0:
                    raise ValueError("Cannot accept a list of 'params' with different lists of filters")

    def get_case(self, identifier, version, as_data_frame=False, include_all=True, **params):
        """
        :type as_data_frame: bool
        :type identifier: str
        :type version: str
        :type include_all: bool
        :rtype: dict | pd.DataFrame
        """
        if include_all:
            params['include'] = [self._INCLUDE_ALL]
        results, _ = self._get("{endpoint}/{identifier}/{version}".format(
            endpoint=self._BASE_ENDPOINT, identifier=identifier, version=version), **params)
        if not results:
            logging.warning("No case found with id-version {}-{}".format(identifier, version))
            return None
        assert len(results) == 1, "Unexpected number of cases returned when searching by identifier"
        return self._render_single_result(results, as_data_frame=as_data_frame)

    def get_case_by_identifiers(self, identifiers, as_data_frame=False, include_all=True, **params):
        """
        :type as_data_frame: bool
        :type identifiers: list
        :type include_all: bool
        :rtype: list | pd.DataFrame
        """
        if include_all:
            params['include'] = [self._INCLUDE_ALL]
        results, _ = self._get("{endpoint}/{identifiers}".format(
            endpoint=self._BASE_ENDPOINT, identifiers=",".join(identifiers)), **params)
        return self._render(results, as_data_frame=as_data_frame)

    def search(self, query):
        results, _ = self._get("{endpoint}/search/{query}".format(endpoint=self._BASE_ENDPOINT, query=query))
        return self._render(results, as_data_frame=False)

    def get_similar_cases_by_case(self, case_id, case_version, as_data_frame=False, **params):
        """
        :type as_data_frame: bool
        :type case_id: str
        :type case_version: int
        :type params: dict
        :rtype: list | pd.DataFrame
        """
        results, _ = self._get([self._BASE_ENDPOINT, case_id, case_version, "similar-cases"], **params)
        if not results:
            logging.warning("No similar cases found")
            return None
        return self._render(results, as_data_frame=as_data_frame)

    def get_similar_cases_by_phenotypes(self, phenotypes, as_data_frame=False, **params):
        """
        :type as_data_frame: bool
        :type phenotypes: list
        :type params: dict
        :rtype: list | pd.DataFrame
        """
        params['hpoIds'] = phenotypes
        results, _ = self._get([self._BASE_ENDPOINT, "phenotypes", "similar-cases"], **params)
        if not results:
            logging.warning("No similar cases found")
            return None
        return self._render(results, as_data_frame=as_data_frame)

    def get_shared_variants_cases_by_case(self, case_id, case_version, report_event_type, **params):
        """
        :type case_id: str
        :type case_version: int
        :type report_event_type: ReportEventType
        :type limit: int
        :type params: dict
        :rtype: list
        """
        assert report_event_type in REPORT_EVENT_TYPES, \
            "Invalid report event type provided '{}'. Valid values: {}".format(report_event_type, REPORT_EVENT_TYPES)
        params['type'] = report_event_type
        results, _ = self._get([self._BASE_ENDPOINT, case_id, case_version, "shared-variants"], **params)
        if not results:
            logging.warning("No cases sharing {} variants found".format(report_event_type))
            return None
        return results

    def get_shared_genes_cases_by_case(self, case_id, case_version, report_event_type, **params):
        """
        :type case_id: str
        :type case_version: int
        :type report_event_type: ReportEventType
        :type params: dict
        :rtype: list
        """
        assert report_event_type in REPORT_EVENT_TYPES, \
            "Invalid report event type provided '{}'. Valid values: {}".format(report_event_type, REPORT_EVENT_TYPES)
        params['type'] = report_event_type
        results, _ = self._get([self._BASE_ENDPOINT, case_id, case_version, "shared-genes"], **params)
        if not results:
            logging.warning("No cases sharing {} genes found".format(report_event_type))
            return None
        return results

    def get_shared_variants_counts(self, variant_ids, **params):
        """
        :type variant_ids: list
        :type params: dict
        :rtype: list
        """
        variant_coordinates = [v.toJsonDict() for v in self.variants().variant_ids_to_coordinates(variant_ids)]
        results, _ = self._post([self._BASE_ENDPOINT, "shared-variants-counts"], variant_coordinates, **params)
        return results

    def get_phenosim_matrix(self, as_data_frame=False, **params):
        """
        :type as_data_frame: bool
        :rtype: list | pd.DataFrame
        """
        results, _ = self._get("{endpoint}/similarity-matrix".format(endpoint=self._BASE_ENDPOINT), **params)
        if not results:
            logging.warning("No similarity matrix found")
            return None
        return self._render(results, as_data_frame=as_data_frame)

    def get_pedigree(self, identifier, version, as_data_frame=False):
        results, _ = self._get("{endpoint}/{identifier}/{version}".format(
            endpoint="pedigrees", identifier=identifier, version=version))
        if not results:
            logging.warning("No pedigree found with id-version {}-{}".format(identifier, version))
            return None
        assert len(results) == 1, "Unexpected number of pedigrees returned when searching by identifier"
        return self._render_single_result(results, as_data_frame=as_data_frame)

    def get_clinical_report(self, identifier, version, as_data_frame=False):
        results, _ = self._get("{endpoint}/{identifier}/{version}".format(
            endpoint="clinical-reports", identifier=identifier, version=version))
        if not results:
            logging.warning("No clinical report found with id-version {}-{}".format(identifier, version))
            return None
        assert len(results) == 1, "Unexpected number of clinical reports returned when searching by identifier"
        return self._render_single_result(results, as_data_frame=as_data_frame)

    def get_rd_exit_questionnaire(self, identifier, version, as_data_frame=False):
        results, _ = self._get("{endpoint}/{identifier}/{version}".format(
            endpoint="rare-disease-exit-questionnaires", identifier=identifier, version=version))
        if not results:
            logging.warning("No questionnaire found with id-version {}-{}".format(identifier, version))
            return None
        assert len(results) == 1, "Unexpected number of questionnaires returned when searching by identifier"
        return self._render_single_result(results, as_data_frame=as_data_frame)

    def get_cancer_participant(self, identifier, version, as_data_frame=False):
        results, _ = self._get("{endpoint}/{identifier}/{version}".format(
            endpoint="participants", identifier=identifier, version=version))
        if not results:
            logging.warning("No cancer participant found with id-version {}-{}".format(identifier, version))
            return None
        assert len(results) == 1, "Unexpected number of cancer participants returned when searching by identifier"
        return self._render_single_result(results, as_data_frame=as_data_frame)



import logging
import os
import random
import uuid
from unittest import TestCase

import pandas as pd
from mock import patch
from protocols.protocol_7_2.cva import Assembly, PedigreeInjectRD, CancerParticipantInject, \
    EvidenceEntryAndVariants, EvidenceEntry, Property, EvidenceSource, Actions, Therapy, DrugResponse, GenomicFeature, \
    FeatureTypes, VariantCoordinates, VariantsCoordinates, Penetrance, DrugResponseClassification, Transaction, \
    TransactionStatus
from protocols.protocol_7_2.reports import Program
from protocols.util import dependency_manager
from protocols.util.factories.avro_factory import GenericFactoryAvro
from requests import ConnectionError

from pyark.cva_client import CvaClient
from pyark.errors import CvaClientError, CvaServerError
from pyark.models.wrappers import ReportEventEntryWrapper, VariantWrapper


class TestPyArk (TestCase):
    # credentials
    CVA_URL_BASE = os.getenv("CVA_URL")
    GEL_USER = os.getenv("CVA_USER")
    GEL_PASSWORD = os.getenv("CVA_PASSWORD")

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(level=logging.DEBUG)
        if cls.GEL_PASSWORD is None:
            cls.GEL_PASSWORD = ""
        if not cls.CVA_URL_BASE or not cls.GEL_USER:
            logging.error("Please set the configuration environment variables: CVA_URL, GEL_USER, GEL_PASSWORD")
            raise ValueError("Missing config")
        logging.info("Running tests against {}".format(cls.CVA_URL_BASE))
        cls.cva = CvaClient(cls.CVA_URL_BASE, user=cls.GEL_USER, password=cls.GEL_PASSWORD, retries=10)
        cls.report_events = cls.cva.report_events()
        cls.entities = cls.cva.entities()
        cls.cases = cls.cva.cases()
        cls.variants = cls.cva.variants()
        cls.data_intake = cls.cva.data_intake()
        # fetch 50 cases to run tests on
        cls.random_cases = list(cls.cases.get_cases(
            max_results=50, program=Program.rare_disease, assembly=Assembly.GRCh38,
            filter='countInterpretationServices.exomiser gt 0 and countTiered gt 100',
            hasClinicalData=True))
        cls.random_closed_cases = list(cls.cases.get_cases(
            max_results=50, program=Program.rare_disease, assembly=Assembly.GRCh38,
            filter='countInterpretationServices.exomiser gt 0 and countTiered gt 100',
            hasClinicalData=True, hasPositiveDx=True))

    def _get_random_case_id_and_version(self, closed=False):
        case = self._get_random_case(closed=closed)
        return case['identifier'], case['version']

    def _get_random_case(self, closed=False):
        return random.choice(self.random_closed_cases if closed else self.random_cases)

    def _get_random_panel(self):
        return random.choice(self._get_random_case()['reportEventsAnalysisPanels'])['panelName']

    def _get_random_gene(self):
        return random.choice(self._get_random_case()['genes'])

    def _get_random_variant_ids(self, n=1):
        return random.sample(self._get_random_case()['allVariants'], n)


class TestReportEvents(TestPyArk):

    def test_get_report_events(self):
        random_case = self._get_random_case()
        page_size = 2
        maximum = 5
        report_events_iterator = self.report_events.get_report_events(
            limit=page_size, max_results=maximum, caseId=random_case['identifier'], caseVersion=random_case['version'])
        re_count = 0
        for re in report_events_iterator:
            self.assertIsNotNone(re)
            self.assertTrue(type(re) == ReportEventEntryWrapper)
            re_count += 1
        self.assertEqual(re_count, maximum)

    def test_count_report_events(self):
        random_case = self._get_random_case()
        count = self.report_events.count(caseId=random_case['identifier'], caseVersion=random_case['version'])
        self.assertIsInstance(count, int)

    def test_variant_summary_by_ids(self):
        variant_ids = self._get_random_variant_ids(n=10)
        results = self.report_events.get_variant_summary_by_ids(variant_ids=variant_ids)
        self.assertIsInstance(results, list)
        self.assertTrue(len(results) == 10)
        for r in results:
            self.assertTrue(r['countCases'] > 0)
            self.assertTrue(r['countCasesClassifiedByAcmg'] is not None)
            self.assertTrue(r['variantId'] in variant_ids)
            self.assertTrue(r['variantCoordinatesGRCh38'] is not None)
            self.assertTrue(r.get('countCasesRareDisease', None) is None)

        results2 = self.report_events.get_variant_summary_by_ids(variant_ids=variant_ids, minimal=True)
        for r in results2:
            self.assertTrue(r['countCases'] > 0)
            self.assertTrue(r['countCasesClassifiedByAcmg'] is not None)
            self.assertTrue(r['variantId'] in variant_ids)
            self.assertTrue(r['variantCoordinatesGRCh38'] is not None)
            self.assertTrue(r.get('countCasesRareDisease', None) is None)

        results3 = self.report_events.get_variant_summary_by_ids(variant_ids=variant_ids, minimal=False)
        for r in results3:
            self.assertTrue(r['countCases'] > 0)
            self.assertTrue(r['countCasesClassifiedByAcmg'] is not None)
            self.assertTrue(r['variantId'] in variant_ids)
            self.assertTrue(r['variantCoordinatesGRCh38'] is not None)
            self.assertTrue(r['countCasesRareDisease'] > 0)

    def test_variant_summary_by_coordinates(self):
        variant_ids = self._get_random_variant_ids(n=10)
        results_by_ids = self.report_events.get_variant_summary_by_ids(variant_ids=variant_ids)
        variant_coordinates = [c.toJsonDict() for c in self.variants.variant_ids_to_coordinates(variant_ids)]
        results_by_coordinates = self.report_events.get_variant_summary_by_coordinates(
            variant_coordinates=variant_coordinates)
        variant_ids_1 = set([r['variantId'] for r in results_by_ids])
        variant_ids_2 = set([r['variantId'] for r in results_by_coordinates])
        self.assertEqual(len(variant_ids_1), len(variant_ids_1.intersection(variant_ids_2)))


class TestCases(TestPyArk):

    def test_get_case(self):

        case_id, case_version = self._get_random_case_id_and_version()

        # gets case
        result = self.cases.get_case(case_id, case_version)
        self.assertTrue(result is not None)
        self.assertTrue(isinstance(result, dict))

        # gets pedigree
        result = self.cases.get_case(case_id, case_version, as_data_frame=True)
        self.assertTrue(result is not None)
        self.assertTrue(isinstance(result, pd.DataFrame))

    def test_get_cases_summary(self):

        summary = self.cases.get_summary(program=Program.rare_disease)
        self.assertIsNotNone(summary)
        self.assertIsInstance(summary, dict)

        summary = self.cases.get_summary(program=Program.rare_disease, as_data_frame=True)
        self.assertIsNotNone(summary)
        self.assertIsInstance(summary, pd.DataFrame)

    def test_get_cases_summary_with_many_queries(self):

        queries = [{'assembly': a} for a in [Assembly.GRCh37, Assembly.GRCh38]]
        summary = self.cases.get_summary(params_list=queries)
        self.assertIsNotNone(summary)
        self.assertIsInstance(summary, list)

        summary = self.cases.get_summary(params_list=queries, as_data_frame=True)
        self.assertIsNotNone(summary)
        self.assertIsInstance(summary, pd.DataFrame)

        try:
            self.cases.get_summary(params_list=queries, program=Program.rare_disease)
            self.assertTrue(False)
        except:
            self.assertTrue(True)

    def test_get_cases(self):

        page_size = 2
        maximum = 5
        cases_iterator = self.cases.get_cases(as_data_frame=False, limit=page_size, max_results=maximum)
        cases_count = 0
        for c in cases_iterator:
            self.assertIsNotNone(c)
            self.assertIsInstance(c, dict)
            cases_count += 1
        self.assertEqual(cases_count, maximum)

    def test_get_cases_as_dataframes(self):

        page_size = 2
        maximum = 5
        cases_iterator = self.cases.get_cases(as_data_frame=True, limit=page_size, max_results=maximum)
        cases_count = 0
        for cases_data_frame in cases_iterator:
            self.assertIsNotNone(cases_data_frame)
            self.assertIsInstance(cases_data_frame, pd.DataFrame)
            cases_count += cases_data_frame.shape[0]
        self.assertEqual(cases_count, maximum)

    def test_get_cases_ids(self):

        page_size = 2
        maximum = 5
        cases_iterator = self.cases.get_cases_ids(as_data_frame=False, limit=page_size, max_results=maximum)
        cases_count = 0
        for c in cases_iterator:
            self.assertIsNotNone(c)
            self.assertIsInstance(c, str)
            cases_count += 1
        self.assertEqual(cases_count, maximum)

    def test_count_cases(self):

        count = self.cases.count()
        self.assertIsInstance(count, int)

    def test_get_similar_cases_by_case(self):

        case_id, case_version = self._get_random_case_id_and_version()

        results = self.cases.get_similar_cases_by_case(case_id=case_id, case_version=case_version)
        if results:
            self.assertIsInstance(results, list)

        results = self.cases.get_similar_cases_by_case(case_id=case_id, case_version=case_version, limit=5)
        if results:
            self.assertIsInstance(results, list)
            self.assertTrue(len(results) <= 5)

    def test_get_similar_cases_by_phenotypes(self):

        phenotypes = ["HP:0000006", "HP:0003186", "HP:0002365"]

        results = self.cases.get_similar_cases_by_phenotypes(phenotypes=phenotypes)
        self.assertIsNotNone(results)
        self.assertIsInstance(results, list)

        results = self.cases.get_similar_cases_by_phenotypes(phenotypes=phenotypes, limit=5)
        self.assertIsNotNone(results)
        self.assertIsInstance(results, list)
        self.assertTrue(len(results) == 5)


class TestVariants(TestPyArk):

    def test_count_variants(self):

        count = self.variants.count()
        self.assertIsInstance(count, int)

    def test_get_variants(self):
        page_size = 2
        maximum = 5
        variants_iterator = self.variants.get_variants(
            limit=page_size, max_results=maximum, genes=[self._get_random_gene()])
        re_count = 0
        for v in variants_iterator:
            self.assertIsNotNone(v)
            self.assertTrue(type(v) == VariantWrapper)
            re_count += 1
        self.assertEqual(re_count, maximum)

    def test_get_variant_by_id(self):

        variant = self.variants.get_variant_by_id(identifier=self._get_random_variant_ids()[0])
        self.assertIsNotNone(variant)
        self.assertIsInstance(variant, VariantWrapper)

    def test_unexisting_variant_by_id(self):
        with self.assertRaises(CvaClientError):
            self.variants.get_variant_by_id(identifier='whatever')

    def test_get_variants_by_id(self):

        identifiers = self._get_random_variant_ids()
        variants = self.variants.get_variants_by_id(identifiers=identifiers)
        self.assertIsNotNone(variants)
        self.assertIsInstance(variants, list)
        for v in variants:
            self.assertIsInstance(v, VariantWrapper)
        self.assertTrue(len(variants) == len(identifiers))

    def test_dont_get_variants_by_id(self):
        non_existing_identifiers = ['whatever', 'this', 'that']
        self.assertRaises(CvaClientError,
                          lambda: self.variants.get_variants_by_id(identifiers=non_existing_identifiers))

    def test_variant_coordinates_to_ids(self):
        expected = self._get_random_variant_ids(n=10)
        variant_coordinates = [c.toJsonDict() for c in self.variants.variant_ids_to_coordinates(expected)]
        observed = self.variants.variant_coordinates_to_ids(variant_coordinates=variant_coordinates)
        self.assertTrue(len(observed) == len(set(observed).intersection(set(expected))))


class TestOthers(TestPyArk):

    def test_get_pedigree(self):

        case_id, case_version = self._get_random_case_id_and_version()

        # gets pedigree
        result = self.cases.get_pedigree(case_id, case_version)
        self.assertTrue(result is not None)
        self.assertTrue(isinstance(result, dict))

        # gets pedigree
        result = self.cases.get_pedigree(case_id, case_version, as_data_frame=True)
        self.assertTrue(result is not None)
        self.assertTrue(isinstance(result, pd.DataFrame))

    def test_get_clinical_report(self):

        case_id, case_version = self._get_random_case_id_and_version(closed=True)

        # gets pedigree
        result = self.cases.get_clinical_report(case_id, case_version)
        self.assertTrue(result is not None)
        self.assertTrue(isinstance(result, dict))

        # gets pedigree
        result = self.cases.get_clinical_report(case_id, case_version, as_data_frame=True)
        self.assertTrue(result is not None)
        self.assertTrue(isinstance(result, pd.DataFrame))

    def test_get_rd_exit_questionnaire(self):

        case_id, case_version = self._get_random_case_id_and_version(closed=True)

        # gets pedigree
        result = self.cases.get_rd_exit_questionnaire(case_id, case_version)
        self.assertTrue(result is not None)
        self.assertTrue(isinstance(result, dict))

        # gets pedigree
        result = self.cases.get_rd_exit_questionnaire(case_id, case_version, as_data_frame=True)
        self.assertTrue(result is not None)
        self.assertTrue(isinstance(result, pd.DataFrame))

    def test_get_all_panels(self):
        panels = self.entities.get_all_panels()
        self.assertIsNotNone(panels)
        self.assertIsInstance(panels, pd.Series)

    def test_get_similarity_matrix(self):

        matrix = self.cases.get_phenosim_matrix(program=Program.rare_disease, specificDiseases='cakut')
        self.assertIsNotNone(matrix)
        self.assertIsInstance(matrix, list)

        matrix = self.cases.get_phenosim_matrix(program=Program.rare_disease, specificDiseases='cakut',
                                                   as_data_frame=True)
        self.assertIsNotNone(matrix)
        self.assertIsInstance(matrix, pd.DataFrame)

    def test_get_panel_summary(self):

        panels = self.entities.get_panels_summary(program=Program.rare_disease)
        self.assertIsNotNone(panels)
        self.assertIsInstance(panels, list)

        panels = self.entities.get_panels_summary(program=Program.cancer, as_data_frame=True)
        self.assertIsNotNone(panels)
        self.assertIsInstance(panels, pd.DataFrame)

    def test_get_disorder_summary(self):

        disorders = self.entities.get_disorders_summary(program=Program.rare_disease)
        self.assertIsNotNone(disorders)
        self.assertIsInstance(disorders, list)

        disorders = self.entities.get_disorders_summary(program=Program.cancer, as_data_frame=True)
        self.assertIsNotNone(disorders)
        self.assertIsInstance(disorders, pd.DataFrame)

    def test_get_gene_summary(self):

        genes = self.entities.get_genes_summary(program=Program.rare_disease)
        self.assertIsNotNone(genes)
        self.assertIsInstance(genes, list)

        genes = self.entities.get_genes_summary(program=Program.cancer, as_data_frame=True)
        self.assertIsNotNone(genes)
        self.assertIsInstance(genes, pd.DataFrame)

    def test_get_phenotypes_summary(self):

        phenotypes = self.entities.get_phenotypes(program=Program.rare_disease)
        self.assertIsNotNone(phenotypes)
        self.assertIsInstance(phenotypes, list)

        phenotypes = self.entities.get_phenotypes(program=Program.cancer, as_data_frame=True)
        self.assertIsNotNone(phenotypes)
        self.assertIsInstance(phenotypes, pd.DataFrame)

    # def test_get_shared_variants_cases(self):
    #     case_id, case_version = self._get_random_case_id_and_version()
    #
    #     results1 = self.cases.get_shared_variants_cases_by_case(
    #         case_id=case_id, case_version=case_version, report_event_type=ReportEventType.genomics_england_tiering)
    #
    #     results2 = self.cases.get_shared_variants_cases_by_case(
    #         case_id=case_id, case_version=case_version, report_event_type=ReportEventType.reported)
    #
    #     results3 = self.cases.get_shared_variants_cases_by_case(
    #         case_id=case_id, case_version=case_version, report_event_type=ReportEventType.questionnaire)
    #
    #     non_null = (r for r in (results1, results2, results3) if r)
    #     self.assertIsNotNone([s for r in non_null for s in r])

    # def test_get_shared_gene_cases(self):
    #     case_id, case_version = self._get_random_case_id_and_version()
    #
    #     results = self.cases.get_shared_genes_cases_by_case(
    #         case_id=case_id, case_version=case_version, report_event_type=ReportEventType.genomics_england_tiering)
    #     self.assertIsNotNone(results)
    #     self.assertIsInstance(results, list)
    #
    #     results = self.cases.get_shared_genes_cases_by_case(
    #         case_id=case_id, case_version=case_version, report_event_type=ReportEventType.reported)
    #     self.assertIsNotNone(results)
    #     self.assertIsInstance(results, list)
    #
    #     results = self.cases.get_shared_genes_cases_by_case(
    #         case_id=case_id, case_version=case_version, report_event_type=ReportEventType.questionnaire)
    #     # NOTE: there are no values
    #     # self.assertIsNotNone(results)
    #     # self.assertIsInstance(results, list)

    def test_get_shared_variants_counts(self):
        case = self._get_random_case()
        all_variants = case['allVariants']
        results = self.cases.get_shared_variants_counts(all_variants)
        for r in results:
            self.assertTrue(r['variantId'] in all_variants)
            self.assertTrue(r['countCases'] >= 1)

    def test_post_pedigree(self):
        transaction = self._test_post(PedigreeInjectRD, self.data_intake.post_pedigree)
        self.assertTrue(isinstance(transaction, Transaction))
        self.assertTrue(transaction.id is not None)
        self.assertTrue(transaction.status == TransactionStatus.PENDING)
        self.assertTrue(transaction.compressedData is None)

    def test_post_participant(self):
        transaction = self._test_post(CancerParticipantInject, self.data_intake.post_participant)
        self.assertTrue(isinstance(transaction, Transaction))
        self.assertTrue(transaction.id is not None)
        self.assertTrue(transaction.status == TransactionStatus.PENDING)
        self.assertTrue(transaction.compressedData is None)

    def test_get_transaction_fails_if_no_results(self):
        # NOTE: this will work when backend returns 404 on this one
        client = self.cva.transactions()
        self.assertRaises(
            CvaClientError,
            lambda: client.get_transaction("notreal")
        )

    def test_errors_if_cva_down(self):
        self.assertRaises(
            ConnectionError,
            lambda: CvaClient("https://nowhere.invalid", user='u', password='p', retries=2).entities().get_all_panels()
        )

    @patch('requests.sessions.Session.get')
    @patch('requests.sessions.Session.post')
    def test_errors_if_4xx(self, post, get):
        self._mock_panels_to_return(get, post, 400)
        self.assertRaises(
            CvaClientError,
            lambda: CvaClient("https://nowhere.invalid", user='u', password='p').entities().get_all_panels()
        )

    @patch('requests.sessions.Session.get')
    @patch('requests.sessions.Session.post')
    def test_errors_if_5xx(self, post, get):
        self._mock_panels_to_return(get, post, 500)
        self.assertRaises(
            CvaServerError,
            lambda: CvaClient("https://nowhere.invalid", user='u', password='p').entities().get_all_panels()
        )

    @patch('requests.sessions.Session.get')
    @patch('requests.sessions.Session.post')
    def test_returns_empty_if_no_results(self, post, get):
        self._mock_panels_to_return(get, post, 200)
        self.assertEqual(0,
                         CvaClient("https://nowhere.invalid", user='u', password='p').entities().get_all_panels().size)

    def test_gets_evidence(self):
        model = create_example_evidence()
        model.evidenceEntry.source.name = str(uuid.uuid1())
        model.evidenceEntry.source.version = str(uuid.uuid1())

        self.cva.evidences().post_evidences(model)

        evidences = self.cva.evidences().get_evidences(
            model.evidenceEntry.source.name,
            version=model.evidenceEntry.source.version
        )
        results = list(evidences)
        self.assertTrue(results, "expected some evidence")

        for item in results:
            self.assertEquals(EvidenceEntryAndVariants, type(item))
            # test random field that has no normalisation against it.
            self.assertEquals(item.evidenceEntry.ethnicity, model.evidenceEntry.ethnicity)

    def test_deleting_case(self):
        transaction = self._test_post(CancerParticipantInject, self.data_intake.post_participant)
        metadata = transaction.transactionDetails.metadata
        caseId = metadata.caseId
        caseVersion = metadata.caseVersion

        transaction = self.cases.delete(caseId, caseVersion)
        self.assertTrue(isinstance(transaction, Transaction))
        self.assertTrue(transaction.id is not None)
        self.assertTrue(transaction.status == TransactionStatus.PENDING)
        self.assertTrue(transaction.compressedData is None)

    @staticmethod
    def _mock_panels_to_return(get, post, status_code):
        auth_response = MockResponse(status_code, {'response': [{'result': [{'token': 'xyz'}]}]})
        post.return_value = auth_response
        response = MockResponse(status_code, {})
        get.return_value = response

    def _test_post(self, clazz, post_function):
        model = GenericFactoryAvro.get_factory_avro(
            clazz=clazz,
            version=dependency_manager.VERSION_70,
            fill_nullables=False,
        ).create()

        response = post_function(model)
        # this is stronger than it looks because post checks for errors
        self.assertIsNotNone(response)
        return response


class MockResponse:
    def __init__(self, status_code, json_dict):
        self.status_code = status_code
        self.json_dict = json_dict
        self.content = str(json_dict)
        self.text = str(json_dict)
        self.headers = {}

    @staticmethod
    def get(key, default):
        return None

    def json(self):
        return self.json_dict


def create_example_evidence():
    coordinates = coordinate()
    marker_coordinates = marker_coordinate()

    genomic_feature = GenomicFeature(
        featureType=FeatureTypes.transcript,
        ensemblId="ENST00000370192.7"
    )

    actions = action()

    source = EvidenceSource(
        date="today",
        name="genomics-england-phamacogenomics",
        version="v1.0"
    )

    pharmgkb_id = Property(
        name="PharmGKB",
        value="PA166153760"
    )

    return EvidenceEntryAndVariants(
        evidenceEntry=EvidenceEntry(
            source=source,
            ethnicity='Z',
            genomicFeatures=[genomic_feature],
            heritable_trait=[],
            penetrance=Penetrance.complete,
            additionalProperties=[pharmgkb_id]
        ),
        variantsCoordinates=VariantsCoordinates(variants=[coordinates]),
        markersCoordinates=VariantsCoordinates(variants=[marker_coordinates]),
        actions=actions
    )


def action():
    return Actions(
        therapies=[
            Therapy(
                drugResponse=[DrugResponse(
                    TreatmentAgent="the drug",
                    drugResponseClassification=DrugResponseClassification.increased_monitoring
                )],
                variantActionable=True,
                referenceUrl="https://something.com/db.html",
                conditions=[]
            )
        ]
    )


def marker_coordinate():
    marker_coordinates = VariantCoordinates(
        assembly=Assembly.GRCh38, reference='G', alternate='T', chromosome='1', position=99450058
    )
    return marker_coordinates


def coordinate():
    coordinates = VariantCoordinates(
        assembly=Assembly.GRCh38, reference='C', alternate='T', chromosome='1', position=97450058
    )
    return coordinates



import re
from protocols.protocol_7_2.cva import Variant, Assembly, VariantRepresentation, \
    VariantAnnotation, ReportEventEntry


class ReportEventEntryWrapper(ReportEventEntry):

    def get_variant(self):
        return VariantWrapper.fromJsonDict(self.observedVariants[0].variant.toJsonDict())


class VariantWrapper(Variant):

    def get_variant_representation_by_assembly(self, assembly):
        """
        :type assembly: Assembly
        :rtype: VariantRepresentation
        """
        for v in self.variants:
            if v.assembly == assembly:
                return v
        return None

    def get_default_variant_representation(self):
        """
        :rtype: VariantRepresentation
        """
        variant_representation = self.get_variant_representation_by_assembly(Assembly.GRCh38)
        if not variant_representation:
            variant_representation = self.get_variant_representation_by_assembly(Assembly.GRCh37)
        return variant_representation

    def get_variant_annotation_by_assembly(self, assembly):
        """
        :type assembly: Assembly
        :rtype: VariantAnnotationWrapper
        """
        variant_representation = self.get_variant_representation_by_assembly(assembly)
        return VariantAnnotationWrapper.fromJsonDict(variant_representation.annotation.toJsonDict())

    def get_default_variant_annotation(self):
        """
        :rtype: VariantAnnotationWrapper
        """
        variant_representation = self.get_default_variant_representation()
        return VariantAnnotationWrapper.fromJsonDict(variant_representation.annotation.toJsonDict())

    def is_small_variant(self):
        """
        :rtype: bool
        """
        return self.get_default_variant_representation().smallVariantCoordinates is not None


class VariantAnnotationWrapper(VariantAnnotation):

    @staticmethod
    def _include_frequency(freq, studies_populations):
        if not studies_populations:
            return True
        freq_dict = {'study': freq.study, 'population': freq.population}
        return freq_dict in studies_populations

    def get_max_allele_frequency(self, studies_populations=[]):
        """
        :type studies_populations: dict
        :rtype: float
        """
        if not self.populationFrequencies:
            return 0.0
        freqs = [freq.altAlleleFreq for freq in self.populationFrequencies
                 if self._include_frequency(freq, studies_populations)]
        if not freqs:
            return 0.0
        return max(freqs)



import re
import logging
import pandas as pd
from pandas.io.json import json_normalize
from pyark.rest_client import RestClient
import multiprocessing


class CvaClient(RestClient):

    _regx = re.compile("^test_", re.IGNORECASE)

    _ENDPOINT_BASE = "cva/api/0"

    _LIMIT_PARAM = 'limit'
    _MARKER_PARAM = 'marker'
    _LIMIT_HEADER = 'X-Pagination-Limit'
    _MARKER_HEADER = 'X-Pagination-Marker'

    # authentication endpoint
    _AUTHENTICATION_ENDPOINT = "authentication"

    _INCLUDE_ALL = "__all"

    def __init__(self, url_base, token=None, user=None, password=None,
                 disable_validation=True, disable_annotation=False, retries=10):

        if not (token or (user and password is not None)):
            logging.error("Credentials are required. Either token or user/password.")
            raise ValueError("Missing credentials")
        RestClient.__init__(self, url_base, self._ENDPOINT_BASE, retries=retries)
        self._disable_validation = disable_validation
        self._disable_annotation = disable_annotation
        self._push_data_params = {'disable_validation': self._disable_validation,
                                 'disable_annotation': self._disable_annotation}
        self._token = "Bearer {}".format(token.replace("Bearer ", "")) if token else None
        self._user = user
        self._password = password
        if self._token or (self._user is not None and self._password is not None):
            self._set_authenticated_header()
        # initialise subclients
        self._report_events_client = None
        self._entities_client = None
        self._cases_client = None
        self._variants_client = None
        self._lift_overs_client = None
        self._data_intake_client = None
        self._transactions_client = None
        self._evidences_client = None

    def _get_token(self):
        results, _ = self._post(self._AUTHENTICATION_ENDPOINT, payload={
            'username': self._user,
            'password': self._password
        })
        return "Bearer {}".format(results[0]['token'])

    def _post(self, endpoint, payload, session=True, **params):
        response, headers = super(CvaClient, self)._post(endpoint, payload, session, **params)
        return CvaClient._parse_result(response), CvaClient._build_next_page_params(headers)

    def _get(self, endpoint, session=True, **params):
        response, headers = super(CvaClient, self)._get(endpoint, session, **params)
        return CvaClient._parse_result(response), CvaClient._build_next_page_params(headers)

    def _delete(self, endpoint, **params):
        response, headers = super(CvaClient, self)._delete(endpoint, **params)
        return CvaClient._parse_result(response), CvaClient._build_next_page_params(headers)

    @staticmethod
    def run_parallel_requests(method, parameters):
        """
        :type method: function
        :type parameters: list
        :rtype: object
        """
        pool = multiprocessing.Pool(processes=10)
        results = list(pool.map(method, parameters))
        pool.close()
        pool.join()
        return results

    def report_events(self):
        """

        :return:
        :rtype: ReportEventsClient
        """
        # NOTE: this import needs to be here due to circular imports
        import pyark.subclients.report_events_client
        if self._report_events_client is None:
            # initialise subclients
            self._report_events_client = pyark.subclients.report_events_client.ReportEventsClient(
                self._url_base, self._token)
        return self._report_events_client

    def entities(self):
        """

        :return:
        :rtype: PanelsClient
        """
        # NOTE: this import needs to be here due to circular imports
        import pyark.subclients.entities_client
        if self._entities_client is None:
            # initialise subclients
            self._entities_client = pyark.subclients.entities_client.EntitiesClient(self._url_base, self._token)
        return self._entities_client

    def cases(self):
        """

        :return:
        :rtype: CasesClient
        """
        # NOTE: this import needs to be here due to circular imports
        import pyark.subclients.cases_client
        if self._cases_client is None:
            # initialise subclients
            self._cases_client = pyark.subclients.cases_client.CasesClient(
                self._url_base, self._token)
        return self._cases_client

    def variants(self):
        """

        :return:
        :rtype: VariantsClient
        """
        # NOTE: this import needs to be here due to circular imports
        import pyark.subclients.variants_client
        if self._variants_client is None:
            # initialise subclients
            self._variants_client = pyark.subclients.variants_client.VariantsClient(
                self._url_base, self._token)
        return self._variants_client

    def transactions(self):
        """

        :return:
        :rtype: VariantsClient
        """
        # NOTE: this import needs to be here due to circular imports
        import pyark.subclients.transactions_client
        if self._transactions_client is None:
            # initialise subclients
            self._transactions_client = pyark.subclients.transactions_client.TransactionsClient(
                self._url_base, self._token)
        return self._transactions_client

    def evidences(self):
        import pyark.subclients.evidences_client
        if self._evidences_client is None:
            # initialise subclients
            self._evidences_client = pyark.subclients.evidences_client.EvidencesClient(
                self._url_base, self._token)
        return self._evidences_client

    def lift_overs(self):
        """

        :return:
        :rtype: LiftOverClient
        """
        # NOTE: this import needs to be here due to circular imports
        import pyark.subclients.lift_over_client
        if self._lift_overs_client is None:
            # initialise subclients
            self._lift_overs_client = pyark.subclients.lift_over_client.LiftOverClient(
                self._url_base, self._token)
        return self._lift_overs_client

    def data_intake(self):
        """

        :return:
        :rtype: DataIntakeClient
        """
        # NOTE: this import needs to be here due to circular imports
        import pyark.subclients.data_intake_client
        if self._data_intake_client is None:
            # initialise subclients
            self._data_intake_client = pyark.subclients.data_intake_client.DataIntakeClient(
                self._url_base, self._token)
        return self._data_intake_client

    @staticmethod
    def _build_next_page_params(headers):
        next_page_params = {}
        limit = headers.get(CvaClient._LIMIT_HEADER, None)
        marker = headers.get(CvaClient._MARKER_HEADER, None)
        if marker:
            next_page_params[CvaClient._LIMIT_PARAM] = limit
            next_page_params[CvaClient._MARKER_PARAM] = marker
        return next_page_params

    @staticmethod
    def _parse_result(response):
        logging.info("Response time : {} ms".format(response.get('time', None)))
        error = response.get('error', None)
        if error:
            logging.error(error)
            raise ValueError(error)
        warning = response.get('warning', None)
        if warning:
            logging.warning(warning)
        if 'response' in response and len(response['response']) > 0 and 'result' in response['response'][0]:
            return response['response'][0]['result']
        else:
            return []

    def _render_single_result(self, results, as_data_frame=False, indexes={}):
        if results is None or len(results) == 0:
            return None
        first = results[0]
        return self._render(first, as_data_frame=as_data_frame, indexes=indexes)

    def _render_multiple_results(self, results, as_data_frame=False):
        if results is None or len(results) == 0:
            return None
        if as_data_frame:
            return pd.concat(results)
        else:
            return results

    @staticmethod
    def _render(results, as_data_frame=False, indexes={}):
        if as_data_frame:
            if results:
                df = json_normalize(results)
                if indexes:
                    df.index = pd.MultiIndex.from_arrays(
                        [[values] if type(values) != list else indexes.values() for values in indexes.values()],
                        names=list(indexes.keys()))
                return df
            else:
                return pd.DataFrame()
        else:
            return results

    def _paginate(self, endpoint, as_data_frame=False, max_results=None, transformer=None, **params):
        more_results = True
        count_returned = 0
        while more_results:
            if max_results and count_returned >= max_results:
                return
            results, next_page_params = self._get(endpoint, **params)
            results = list(results)
            if transformer:
                results = list(map(transformer, results))
            if next_page_params:
                params[CvaClient._LIMIT_PARAM] = next_page_params[CvaClient._LIMIT_PARAM]
                params[CvaClient._MARKER_PARAM] = next_page_params[CvaClient._MARKER_PARAM]
            else:
                more_results = False
            if max_results and len(results) > max_results - count_returned:
                # removes those elements in the page that overflow the maximum parameter
                results = results[0:max_results-count_returned]
            # NOTE: when returning a data frame we want all results in a batch in the
            # same data frame, otherwise we want to iterate through them one by one
            if as_data_frame:
                df = self._render(results, as_data_frame=as_data_frame)
                df['_index'] = list(range(count_returned, count_returned + len(results)))
                df.set_index('_index', drop=True, inplace=True)
                count_returned += len(results)
                yield df
            else:
                for r in results:
                    count_returned += 1
                    yield r



import re
import logging
import pyark.cva_client as cva_client
from pyark.models.wrappers import VariantWrapper
from protocols.protocol_7_2.cva import VariantCoordinates


_singleton_instance = None


# NOTE: this method needs to be out of any class as it needs to be pickled (ie: serialised) by multiprocessing library
def _get_variant_by_id(identifier):
    return _singleton_instance.get_variant_by_id(identifier)


class VariantsClient(cva_client.CvaClient):

    _BASE_ENDPOINT = "variants"

    def __init__(self, url_base, token):
        cva_client.CvaClient.__init__(self, url_base, token=token)

    def count(self, **params):
        """
        :type params: dict
        :rtype: int
        """
        return self.get_variants(count=True, **params)

    def get_variant_by_id(self, identifier, include_all=True, **params):
        """
        :type identifier: str
        :type include_all: bool
        :rtype: VariantWrapper
        """
        if include_all:
            params['include'] = [self._INCLUDE_ALL]
        results, _ = self._get("{endpoint}/{identifier}".format(
            endpoint=self._BASE_ENDPOINT, identifier=identifier), **params)
        if not results:
            logging.warning("No variant found with id {}".format(identifier))
            return None
        assert len(results) == 1, "Unexpected number of variants returned when searching by identifier"
        variant = VariantWrapper.fromJsonDict(results[0])
        return variant

    def get_variants_by_id(self, identifiers):
        """
        :type identifiers: list
        :rtype: list
        """
        self._set_singleton()
        return VariantsClient.run_parallel_requests(_get_variant_by_id, identifiers)

    def get_variants(self, as_data_frame=False, max_results=None, include_all=True, **params):
        """
        :type as_data_frame: bool
        :type max_results: int
        :type include_all: bool
        :type params: dict
        :rtype: generator
        """
        if params.get('count', False):
            results, next_page_params = self._get(self._BASE_ENDPOINT, **params)
            return results[0]
        else:
            if include_all:
                params['include'] = [self._INCLUDE_ALL]
            if not as_data_frame:
                def transformer(x): return VariantWrapper.fromJsonDict(x)
            else:
                transformer = None
            return self._paginate(
                endpoint=self._BASE_ENDPOINT, as_data_frame=as_data_frame, max_results=max_results,
                transformer=transformer, **params)

    def variant_ids_to_coordinates(self, variant_ids, fail_on_structural=False):
        """
        :type variant_ids: list
        :type fail_on_structural: bool
        :rtype: list
        """
        return list(filter(lambda x: x is not None,
                           [self.variant_id_to_coordinates(v, fail_on_structural) for v in variant_ids]))

    def variant_id_to_coordinates(self, variant_id, fail_on_structural=False):
        """
        :type variant_id: str
        :type fail_on_structural: bool
        :rtype: VariantCoordinates
        """
        match = re.match(r'(GRCh37|GRCh38):(.+):([ 0-9]+):(-| |[A|C|G|T]*):(-| |[A|C|G|T]*)', variant_id)
        if match and (len(match.group(4)) > 0 or len(match.group(5)) > 0):
            variant_coordinates = VariantCoordinates.fromJsonDict({
                'assembly': match.group(1),
                'chromosome': match.group(2),
                'position': match.group(3).strip(),
                'reference': match.group(4),
                'alternate': match.group(5)
            })
        else:
            variant = self.get_variant_by_id(variant_id)
            if variant is None:
                raise ValueError("The variant id {} could not be mapped".format(variant_id))
            variant_representation = variant.get_default_variant_representation()
            variant_coordinates = variant_representation.smallVariantCoordinates
            if variant_coordinates is None:
                if fail_on_structural:
                    raise ValueError("The variant id {} does not correspond to a small variant".format(variant_id))
                else:
                    return None
        return variant_coordinates

    def variant_coordinates_to_ids(self, variant_coordinates):
        """
        :type variant_coordinates: list
        :rtype: list
        """
        results, _ = self._post([self._BASE_ENDPOINT, "identifiers-from-small-variant-coordinates"], variant_coordinates)
        return results

    def _set_singleton(self):
        global _singleton_instance
        _singleton_instance = self



from pyark import cva_client
from protocols.protocol_7_2.cva import (
    PedigreeInjectRD,
    CancerParticipantInject,
    InterpretedGenomeInject,
    ClinicalReportInject,
    ExitQuestionnaireInjectRD,
    ExitQuestionnaireInjectCancer,
    Transaction
)


class DataIntakeClient(cva_client.CvaClient):

    _INTERPRETED_GENOME_POST = "interpreted-genomes"
    _CLINICAL_REPORT_POST = "clinical-reports"
    _EXIT_QUESTIONAIRES_RD_POST = "exit-questionnaires-rd"
    _EXIT_QUESTIONAIRES_CANCER_POST = "exit-questionnaires-cancer"
    _PEDIGREE_POST = "pedigrees"
    _PARTICIPANT_POST = "participants"

    def __init__(self, url_base, token):
        cva_client.CvaClient.__init__(self, url_base, token=token)

    def post_pedigree(self, pedigree, params={}):
        """
        :type pedigree: PedigreeInjectRD
        :type params: dict
        :rtype: Transaction
        """
        results, _ = self._post(self._PEDIGREE_POST, pedigree.toJsonDict(), params)
        result = self._render_single_result(results, as_data_frame=False)
        return Transaction.fromJsonDict(result) if result else None

    def post_participant(self, participant, params={}):
        """
        :type participant: CancerParticipantInject
        :type params: dict
        :rtype: Transaction
        """
        results, _ = self._post(self._PARTICIPANT_POST, participant.toJsonDict(), params)
        result = self._render_single_result(results, as_data_frame=False)
        return Transaction.fromJsonDict(result) if result else None

    def post_interpreted_genome(self, tiered_variant, params={}):
        """
        :type tiered_variant: InterpretedGenomeInject
        :type params: dict
        :rtype: Transaction
        """
        results, _ = self._post(self._INTERPRETED_GENOME_POST, tiered_variant.toJsonDict(), params)
        result = self._render_single_result(results, as_data_frame=False)
        return Transaction.fromJsonDict(result) if result else None

    def post_clinical_report(self, candidate_variant, params={}):
        """
        :type candidate_variant: ClinicalReportInject
        :type params: dict
        :rtype: Transaction
        """
        results, _ = self._post(self._CLINICAL_REPORT_POST, candidate_variant.toJsonDict(), params)
        result = self._render_single_result(results, as_data_frame=False)
        return Transaction.fromJsonDict(result) if result else None

    def post_exit_questionaire(self, exit_questionaire, params={}):
        """
        :type exit_questionaire: ExitQuestionnaireInjectRD
        :type params: dict
        :rtype: Transaction
        """
        results, _ = self._post(self._EXIT_QUESTIONAIRES_RD_POST, exit_questionaire.toJsonDict(), params)
        result = self._render_single_result(results, as_data_frame=False)
        return Transaction.fromJsonDict(result) if result else None

    def post_exit_questionaire_cancer(self, exit_questionaire, params={}):
        """
        :type exit_questionaire: ExitQuestionnaireInjectCancer
        :type params: dict
        :rtype: Transaction
        """
        results, _ = self._post(self._EXIT_QUESTIONAIRES_CANCER_POST, exit_questionaire.toJsonDict(), params)
        result = self._render_single_result(results, as_data_frame=False)
        return Transaction.fromJsonDict(result) if result else None